#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <errno.h>
#include <string.h>
#include <malloc.h>
#include <sys/syscall.h>
#include <unistd.h>

// ---- Define a (thread-safe) stack here ----

struct stack_element {
  int element;
  struct stack_element* prev;
};

typedef struct {
  struct stack_element* top;
  pthread_mutex_t lock;
} stack;

stack* initialize_stack() {
  stack* nstack = malloc(sizeof(stack));
  if (nstack == NULL) {
    perror("malloc error");
    exit(EXIT_FAILURE);
  }
  nstack->top = NULL;
    
  if (pthread_mutex_init(&nstack->lock, NULL) != 0) {
    fprintf(stderr, "Can't initialize lock!\n");
    exit(EXIT_FAILURE);
  }

  return nstack;
}

void acquire_stack_lock(stack* nstack) {
  int err = pthread_mutex_lock(&nstack->lock);
  if (err != 0) {
    fprintf(stderr, "Can't acquire lock! (%s)\n", strerror(err));
    exit(EXIT_FAILURE);
  }
}

void release_stack_lock(stack* nstack) {
  if (pthread_mutex_unlock(&nstack->lock) != 0) {
    fprintf(stderr, "Can't unlock lock!\n");
    exit(EXIT_FAILURE);
  }
}

int empty(stack* nstack) {
  // TODO - make me thread-safe  
  return nstack->top == NULL;
}

int top(stack* nstack) {
  // TODO - make me thread-safe
  if (empty(nstack)) {
    fprintf(stderr, "top/pop on empty stack\n");
    return -1;
  }

  return nstack->top->element;
}

int pop(stack* nstack) {
  // TODO - make me thread-safe  
  if (empty(nstack)) {
    fprintf(stderr, "top/pop on empty stack\n");
    return -1;
  }

  struct stack_element* top = nstack->top;
  int top_val = top->element;
  nstack->top = nstack->top->prev;
  //  free(top); // possible memory corruption! - uncomment only after making all thread-safe

  return top_val;
}

void push(stack* nstack, int value) {
  // TODO - make me thread-safe  

  struct stack_element* new_element = malloc(sizeof(struct stack_element));
  if (new_element == NULL) {
    perror("malloc error");
    exit(EXIT_FAILURE);
  }

  new_element->element = value;
  new_element->prev = nstack->top;
  nstack->top = new_element;
}

void destroy_stack(stack* nstack) {
  // TODO - make me thread-safe
  while (!empty(nstack)) pop(nstack);
  free(nstack);
  pthread_mutex_destroy(&nstack->lock);  
}

// ---- define thread test operations ----

void* runnable(void* args) {
  int verbose = 0;

  int operations = 100; // run this many push/pop operations
  stack* nstack = (stack*) args;
  int tid = syscall(SYS_gettid);  

  for (int i = 0; i < operations; ++i) {
    // non-atomic, may be misleading
    if (verbose) printf("thread %d: push(%d)\n", tid, i);
    push(nstack,i);
  }
  for (int i = 0; i < operations; ++i) {
    int val = pop(nstack);
    if (verbose) printf("thread %d: popped %d\n", tid, val);
  }

  return NULL;
}

int main(int argc, char** argv) {

  // the main thread: insert some elements, remove them and check that
  // the stack is empty - all should be fine and no memory leaks
  // (check with valgrind)

  printf("-- running the main thread --\n");
  stack* nstack = initialize_stack();

  int operations = 10;
  for (int i = 0; i < operations; ++i) {
    push(nstack,i);
  }

  for (int i = 0; i < operations; ++i) {
    printf("%d\n", pop(nstack));
  }
  if (empty(nstack)) printf("The stack is empty.\n");
  else printf("Nonempty stack!\n");
  destroy_stack(nstack);

  // the multithreaded part: make some threads, make them add
  // some elements to the stack, remove them, and see if the stack is
  // empty (alternatively: look for failures)

  const int pool_size = 10;
  pthread_t thread_pool[pool_size];
  stack* mstack = initialize_stack();

  printf("-- running the multithreaded stack experiment with %d threads --\n", pool_size);

  for (int i = 0; i < pool_size; ++i) pthread_create(&thread_pool[i], NULL, runnable, mstack);

  // join all threads and destroy the stack
  for (int i = 0; i < pool_size; ++i) {
    pthread_join(thread_pool[i], NULL);
  }

  if (empty(mstack)) printf("The stack is empty.\n");
  else printf("Nonempty stack!\n");
  destroy_stack(mstack);

  return 0;
}
